/* bucket.c - The routines for playing with hash buckets. */

/*  This file is part of GDBM, the GNU data base manager, by Philip A. Nelson.
    Copyright (C) 1990  Free Software Foundation, Inc.

    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    GDBM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GDBM; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  phil@wwu.edu
      us-mail:  Philip A. Nelson
                Computer Science Department
                Western Washington University
                Bellingham, WA 98226
        phone:  (206) 676-3035
       
*************************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include "gdbmdefs.h"
#include "systems.h"


/* Initializing a new hash buckets sets all bucket entries to -1 hash value. */
_gdbm_new_bucket (dbf, bucket, bits)
     gdbm_file_info *dbf;
     hash_bucket *bucket;
     int bits;
{
  int index;
  
  /* Initialize the avail block. */
  bucket->av_count = 0;
  for(index = 0 ; index < BUCKET_AVAIL ; index++)
  {
	bucket->bucket_avail[index].av_size = 0 ;
	bucket->bucket_avail[index].av_adr = -1 ;
  }

  /* Set the information fields first. */
  bucket->bucket_bits = bits;
  bucket->count = 0;
  
  /* Initialize all bucket elements. */
  for (index = 0; index < dbf->header->bucket_elems; index++)
  {
    char *cp ;
    bucket->h_table[index].hash_value = -1;
    for(cp = &bucket->h_table[index].key_start[0] ;
	cp < &bucket->h_table[index].key_start[SMALL] ; cp++)
		*cp = 0 ;
    bucket->h_table[index].data_pointer = -1 ;
    bucket->h_table[index].key_size = 0 ;
    bucket->h_table[index].data_size = 0 ;
  }
}


int
_gdbm_read_avail_elem (dbf, av_elem)
     gdbm_file_info *dbf;
     avail_elem *av_elem ;
{
	u_long buf[2] ;

  	if(fread(buf, sizeof(buf), 1, dbf->fp) != 1)
		return -1 ;
	av_elem->av_size = ntohi(buf[0]) ;
	av_elem->av_adr = ntohi(buf[1]) ;
	return 0 ;
}


static
_gdbm_read_bucket_elem (dbf, bu_elem)
     gdbm_file_info *dbf;
     bucket_element *bu_elem ;
{
	u_long buf[3] ;

  	if(fread(buf, sizeof(u_long), 1, dbf->fp) != 1)
    		_gdbm_fatal (dbf, "read error");
	bu_elem->hash_value = ntohi(buf[0]) ;

	/* this is only correct if SMALL is an integer multiple of four */
  	if(fread(&bu_elem->key_start[0], SMALL, 1, dbf->fp) != 1)
    		_gdbm_fatal (dbf, "read error");

  	if(fread(buf, sizeof(buf), 1, dbf->fp) != 1)
    		_gdbm_fatal (dbf, "read error");
	bu_elem->data_pointer = ntohi(buf[0]) ;
	bu_elem->key_size = ntohi(buf[1]) ;
	bu_elem->data_size = ntohi(buf[2]) ;
}



/* Find a bucket for DBF that is pointed to by the bucket directory from
   location DIR_INDEX.   The bucket cache is first checked to see if it
   is already in memory.  If not, a bucket may be tossed to read the new
   bucket.  In any case, the requested bucket is make the "current" bucket
   and dbf->bucket points to the correct bucket. */
int
_gdbm_get_bucket (dbf, dir_index)
     gdbm_file_info *dbf;
     int dir_index;
{
  long  bucket_adr;	/* The address of the correct hash bucket.  */
  int  num_bytes;	/* The number of bytes read. */
  int  index;		/* Loop index. */

  /* Initial set up. */
  dbf->bucket_dir = dir_index;
  bucket_adr = dbf->dir [dir_index];
  
  /* Is that one is not alread current, we must find it. */
  if (dbf->cache_entry->ca_adr != bucket_adr)
    {
      /* Look in the cache. */
      for (index = 0; index < CACHE_SIZE; index++)
	if (dbf->bucket_cache[index].ca_adr == bucket_adr)
	  {
	    dbf->bucket = dbf->bucket_cache[index].ca_bucket;
	    dbf->cache_entry = &dbf->bucket_cache[index];
	    return;
	  }

      /* It is not in the cache, read it from the disk. */
      dbf->last_read = (dbf->last_read + 1) % CACHE_SIZE;
      dbf->bucket_cache[dbf->last_read].ca_adr = bucket_adr;
      dbf->bucket = dbf->bucket_cache[dbf->last_read].ca_bucket;
      dbf->cache_entry = &dbf->bucket_cache[dbf->last_read];
      dbf->cache_entry->ca_data.elem_loc = -1;
      dbf->cache_entry->ca_changed = FALSE;

        /* Read the bucket. */
	{
	hash_bucket *bu = dbf->bucket ;
	u_long buf[2] ;
	int ii ;

	if (fseek (dbf->fp, bucket_adr, L_SET) != 0)
		_gdbm_fatal (dbf, "seek error");

  	if(fread(buf, sizeof(u_long), 1, dbf->fp) != 1)
    		_gdbm_fatal (dbf, "read error");
	bu->av_count = ntohi(buf[0]) ;

	for(ii = 0 ; ii < BUCKET_AVAIL ; ii++)
		if(_gdbm_read_avail_elem (dbf, &bu->bucket_avail[ii])  < 0)
    			_gdbm_fatal (dbf, "read error");

  	if(fread(buf, sizeof(buf), 1, dbf->fp) != 1)
    		_gdbm_fatal (dbf, "read error");
	bu->bucket_bits = ntohi(buf[0]) ;
	bu->count = ntohi(buf[1]) ;

	for(ii = 0 ; ii < dbf->header->bucket_elems ; ii++)
		_gdbm_read_bucket_elem (dbf, &bu->h_table[ii]) ;
	}
    }

  return;
}


/* Split the current bucket.  This includes moving all items in the bucket to
   a new bucket.  This doesn't require any disk reads because all hash values
   are stored in the buckets.  Splitting the current bucket may require
   doubling the size of the hash directory.  */

_gdbm_split_bucket (dbf, next_insert)
     gdbm_file_info *dbf;
     int next_insert;
{
  hash_bucket *bucket[2]; 	/* Pointers to the new buckets. */

  int          new_bits;	/* The number of bits for the new buckets. */
  int	       cache_0;		/* Location in the cache for the buckets. */
  int	       cache_1;
  int	       adr_0;		/* File address of the new bucket 0. */
  int	       adr_1;		/* File address of the new bucket 1. */
  avail_elem   old_bucket;	/* Avail Struct for the old bucket. */

  int          dir_start0;	/* Used in updating the directory. */
  int          dir_start1;
  int          dir_end;

  int	      *new_dir;		/* Pointer to the new directory. */
  int          dir_adr; 	/* Address of the new directory. */
  int          dir_size;	/* Size of the new directory. */
  int          old_adr[31]; 	/* Address of the old directories. */
  int          old_size[31]; 	/* Size of the old directories. */
  int	       old_count;	/* Number of old directories. */

  int          index;		/* Used in array indexing. */
  int          index1;		/* Used in array indexing. */
  int          elem_loc;	/* Location in new bucket to put element. */
  bucket_element *old_el;	/* Pointer into the old bucket. */
  int	       select;		/* Used to index bucket during movement. */


  /* No directories are yet old. */
  old_count = 0;

  while (dbf->bucket->count == dbf->header->bucket_elems)
    {
      /* Initialize the "new" buckets in the cache. */
      do
	{
	  dbf->last_read = (dbf->last_read + 1) % CACHE_SIZE;
	  cache_0 = dbf->last_read;
	}      
      while (dbf->bucket_cache[cache_0].ca_bucket == dbf->bucket);
      bucket[0] = dbf->bucket_cache[cache_0].ca_bucket;
      if (dbf->bucket_cache[cache_0].ca_changed)
	_gdbm_write_cache_elem (dbf, &dbf->bucket_cache[cache_0]);
      do
	{
	  dbf->last_read = (dbf->last_read + 1) % CACHE_SIZE;
	  cache_1 = dbf->last_read;
	}      
      while (dbf->bucket_cache[cache_1].ca_bucket == dbf->bucket);
      bucket[1] = dbf->bucket_cache[cache_1].ca_bucket;
      if (dbf->bucket_cache[cache_1].ca_changed)
	_gdbm_write_cache_elem (dbf, &dbf->bucket_cache[cache_1]);
      new_bits = dbf->bucket->bucket_bits+1;
      _gdbm_new_bucket (dbf, bucket[0], new_bits);
      _gdbm_new_bucket (dbf, bucket[1], new_bits);
      adr_0 = _gdbm_alloc (dbf, dbf->header->bucket_size); 
      dbf->bucket_cache[cache_0].ca_adr = adr_0;
      adr_1 = _gdbm_alloc (dbf, dbf->header->bucket_size);
      dbf->bucket_cache[cache_1].ca_adr = adr_1;

      
      
      /* Double the directory size if necessary. */
      if (dbf->header->dir_bits == dbf->bucket->bucket_bits)
	{
	  dir_size = dbf->header->dir_size * 2;
	  dir_adr  = _gdbm_alloc (dbf, dir_size);
	  new_dir  = (int *) malloc (dir_size);
	  if (new_dir == NULL) _gdbm_fatal (dbf, "malloc error");
	  for (index = 0; index < dbf->header->dir_size/sizeof (int); index++)
	    {
	      new_dir[2*index]   = dbf->dir[index];
	      new_dir[2*index+1] = dbf->dir[index];
	    }
	  
	  /* Update header. */
	  old_adr[old_count] = dbf->header->dir;
	  dbf->header->dir = dir_adr;
	  old_size[old_count] = dbf->header->dir_size;
	  dbf->header->dir_size = dir_size;
	  dbf->header->dir_bits = new_bits;
	  old_count++;
	  
	  /* Now update dbf.  */
	  dbf->header_changed = TRUE;
	  dbf->bucket_dir *= 2;
	  free (dbf->dir);
	  dbf->dir = new_dir;
	}

      /* Copy all elements in dbf->bucket into the new buckets. */
      for (index = 0; index < dbf->header->bucket_elems; index++)
	{
	  old_el = & (dbf->bucket->h_table[index]);
	  select = (old_el->hash_value >> (31-new_bits)) & 1;
	  elem_loc = old_el->hash_value % dbf->header->bucket_elems;
	  while (bucket[select]->h_table[elem_loc].hash_value != -1)
	    elem_loc = (elem_loc + 1) % dbf->header->bucket_elems;
	  bucket[select]->h_table[elem_loc] = *old_el;
	  bucket[select]->count += 1;
	}
      
      /* Allocate avail space for the bucket[1]. */
      bucket[1]->bucket_avail[0].av_adr
	= _gdbm_alloc (dbf, dbf->header->block_size);
      bucket[1]->bucket_avail[0].av_size = dbf->header->block_size;
      bucket[1]->av_count = 1;
      
      /* Copy the avail elements in dbf->bucket to bucket[0]. */
      bucket[0]->av_count = dbf->bucket->av_count;
      index = 0;
      index1 = 0;
      if (bucket[0]->av_count == BUCKET_AVAIL)
	{
	  /* The avail is full, move the first one to bucket[1]. */
	  _gdbm_put_av_elem (dbf->bucket->bucket_avail[0],
			     bucket[1]->bucket_avail,
			     &bucket[1]->av_count);
	  index = 1;
	  bucket[0]->av_count --;
	}
      for (; index < dbf->bucket->av_count; index++)
	{
	  bucket[0]->bucket_avail[index1++] = dbf->bucket->bucket_avail[index];
	}
      
      /* Update the directory.  We have new file addresses for both buckets. */
      dir_start1 = (dbf->bucket_dir >> (dbf->header->dir_bits - new_bits)) | 1;
      dir_end = (dir_start1 + 1) << (dbf->header->dir_bits - new_bits);
      dir_start1 = dir_start1 << (dbf->header->dir_bits - new_bits);
      dir_start0 = dir_start1 - (dir_end - dir_start1);
      for (index = dir_start0; index < dir_start1; index++)
	dbf->dir[index] = adr_0;
      for (index = dir_start1; index < dir_end; index++)
	dbf->dir[index] = adr_1;
      
      
      /* Set changed flags. */
      dbf->bucket_cache[cache_0].ca_changed = TRUE;
      dbf->bucket_cache[cache_1].ca_changed = TRUE;
      dbf->bucket_changed = TRUE;
      dbf->directory_changed = TRUE;
      dbf->second_changed = TRUE;
      
      /* Update the cache! */
      dbf->bucket_dir = next_insert >> (31-dbf->header->dir_bits);
      
      /* Invalidate old cache entry. */
      old_bucket.av_adr  = dbf->cache_entry->ca_adr;
      old_bucket.av_size = dbf->header->bucket_size;
      dbf->cache_entry->ca_adr = 0;
      dbf->cache_entry->ca_changed = FALSE;
      
      /* Set dbf->bucket to the proper bucket. */
      if (dbf->dir[dbf->bucket_dir] == adr_0)
	{
	  dbf->bucket = bucket[0];
	  dbf->cache_entry = &dbf->bucket_cache[cache_0];
	  _gdbm_put_av_elem (old_bucket,
			     bucket[1]->bucket_avail,
			     &bucket[1]->av_count);
	}
      else
	{
	  dbf->bucket = bucket[1];
	  dbf->cache_entry = &dbf->bucket_cache[cache_1];
	  _gdbm_put_av_elem (old_bucket,
			     bucket[0]->bucket_avail,
			     &bucket[0]->av_count);
	}
      
    }

  /* Get rid of old directories. */
  for (index = 0; index < old_count; index++)
    _gdbm_free (dbf, old_adr[index], old_size[index]);
}


int
_gdbm_write_avail_elem (dbf, av_elem)
     gdbm_file_info *dbf;
     avail_elem *av_elem ;
{
	u_long buf[2] ;

	buf[0] = htoni(av_elem->av_size) ;
	buf[1] = htonl(av_elem->av_adr) ;
  	if(fwrite(buf, sizeof(buf), 1, dbf->fp) != 1)
    		return -1 ;
	return 0 ;
}


static int
_gdbm_write_bucket_elem (dbf, bu_elem)
     gdbm_file_info *dbf;
     bucket_element *bu_elem ;
{
	u_long buf[3] ;

	buf[0] = htoni(bu_elem->hash_value) ;
  	if(fwrite(buf, sizeof(u_long), 1, dbf->fp) != 1)
    		return -1 ;

	/* this is only correct if SMALL is an integer multiple of four */
  	if(fwrite(&bu_elem->key_start[0], SMALL, 1, dbf->fp) != 1)
    		return -1 ;

	buf[0] = htonl(bu_elem->data_pointer) ;
	buf[1] = htoni(bu_elem->key_size) ;
	buf[2] = htoni(bu_elem->data_size) ;
  	if(fwrite(buf, sizeof(buf), 1, dbf->fp) != 1)
    		return -1 ;
	return 0 ;
}


/* The only place where a bucket is written. */

int
_gdbm_write_bucket (dbf, bu)
     gdbm_file_info *dbf;
     hash_bucket *bu ;
{
	u_long buf[2] ;
	int ii ;

	buf[0] = htoni(bu->av_count) ;
  	if(fwrite(buf, sizeof(u_long), 1, dbf->fp) != 1)
		return -1 ;

	for(ii = 0 ; ii < BUCKET_AVAIL ; ii++)
		if( _gdbm_write_avail_elem (dbf, &bu->bucket_avail[ii]) < 0)
			return -1 ;

	buf[0] = htoni(bu->bucket_bits) ;
	buf[1] = htoni(bu->count) ;
  	if(fwrite(buf, sizeof(buf), 1, dbf->fp) != 1)
		return -1 ;

	for(ii = 0 ; ii < dbf->header->bucket_elems ; ii++)
		if(_gdbm_write_bucket_elem (dbf, &bu->h_table[ii]) < 0)
			return -1 ;

	return 0 ;
}


/* cache entry containing the bucket to be written. */
_gdbm_write_cache_elem (dbf, ca_entry)
     gdbm_file_info *dbf;
     cache_elem *ca_entry;
{
	hash_bucket *bu = ca_entry->ca_bucket ;
	u_long buf[2] ;
	int ii ;

	if (fseek (dbf->fp, ca_entry->ca_adr, L_SET) != 0)
    		_gdbm_fatal (dbf, "seek error");
		
	if(_gdbm_write_bucket (dbf, ca_entry->ca_bucket) < 0 
			|| fflush(dbf->fp) != 0)
    		_gdbm_fatal (dbf, "write error");

	ca_entry->ca_changed = FALSE;
	ca_entry->ca_data.hash_val = -1;
	ca_entry->ca_data.elem_loc = -1;
}
