/* update.c - The routines for updating the file to a consistent state. */

/*  This file is part of GDBM, the GNU data base manager, by Philip A. Nelson.
    Copyright (C) 1990  Free Software Foundation, Inc.

    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    GDBM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GDBM; see the file COPYING.  If not, write to
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

    You may contact the author by:
       e-mail:  phil@wwu.edu
      us-mail:  Philip A. Nelson
                Computer Science Department
                Western Washington University
                Bellingham, WA 98226
        phone:  (206) 676-3035
       
*************************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include "gdbmdefs.h"
#include "systems.h"


/* This procedure writes the header back to the file described by DBF. */

int
_gdbm_write_header (dbf)
     gdbm_file_info *dbf;
{
	gdbm_file_header *hdr = dbf->header ;
	u_long buf[8] ;
	
	if (fseek (dbf->fp, (long)0, L_SET) != 0)
		return -1 ; /* failure */
		

	buf[0] = htonl(hdr->header_magic) ;
	buf[1] = htonl(hdr->block_size) ;
	buf[2] = htonl(hdr->dir) ;
	buf[3] = htoni(hdr->dir_size) ;
	buf[4] = htoni(hdr->dir_bits) ;
	buf[5] = htoni(hdr->bucket_size) ;
	buf[6] = htoni(hdr->bucket_elems) ;
	buf[7] = htonl(hdr->next_block) ;

	if (fwrite (buf, sizeof(buf), 1, dbf->fp) != 1)
		return -1 ;

	if(_gdbm_write_avail_block(dbf, &hdr->avail) < 0 )
		return -1 ;

	return 0 ;
}


/* Write the directory. */
int
_gdbm_write_dir (dbf)
     gdbm_file_info *dbf;
{
	u_long buf[1] ;
	int ii ;
	int *ip ;
	if (fseek (dbf->fp, dbf->header->dir, L_SET) != 0)
		return -1 ;

	/* improve performance here */
	for(ii = 0, ip = dbf->dir ; ii < dbf->header->dir_size/sizeof(int) ;
		ii++, ip++)
	{
		*buf = htoni(*ip) ;
		if (fwrite (buf, sizeof(buf), 1, dbf->fp) != 1)
			return -1 ;
	}

	return 0 ;
}

/* After all changes have been made in memory, we now write them
   all to disk. */
_gdbm_end_update (dbf)
     gdbm_file_info *dbf;
{
  int num_bytes;	/* Return value for write. */
  
  /* Write the current bucket. */
  if (dbf->bucket_changed)
    {
      _gdbm_write_cache_elem (dbf, dbf->cache_entry);
      dbf->bucket_changed = FALSE;
    }

  /* Write the other changed buckets if there are any. */
  if (dbf->second_changed)
    {
      int index;

      for (index = 0; index < CACHE_SIZE; index++)
	if (dbf->bucket_cache[index].ca_changed)
	  {
	    _gdbm_write_cache_elem (dbf, &dbf->bucket_cache[index]);
	  }
      dbf->second_changed = FALSE;
    }
  
  if (dbf->directory_changed)
  {
	if(_gdbm_write_dir (dbf) < 0
			|| fflush(dbf->fp) != 0)
		_gdbm_fatal (dbf, "write error");

	dbf->directory_changed = FALSE;
	if (!dbf->header_changed)
		(void) fsync (fileno(dbf->fp));
  }

  /* Final write of the header. */
  if (dbf->header_changed)
    {
	if(_gdbm_write_header (dbf)  < 0
		|| fflush(dbf->fp) != 0)
		_gdbm_fatal (dbf, "write error");

	/* Wait for all output to be done. */
	(void)fsync (fileno(dbf->fp));

      dbf->header_changed = FALSE;
    }
}


/* If a fatal error is detected, come here and exit. VAL tells which fatal
   error occured. */
_gdbm_fatal (dbf, val)
     gdbm_file_info *dbf;
     char *val;
{
  if (dbf->fatal_err != NULL)
    (*dbf->fatal_err) (val);
  else
    fprintf (stderr, "gdbm fatal: %s.\n", val);
  exit (-1);
}

